# -*- coding: utf-8 -*-

'''
Settings and configuration for 51Degrees Mobile Detector.

Values will be read from the file or module specified by the
FIFTYONE_DEGREES_MOBILE_DETECTOR_SETTINGS environment variable. Both file
paths (e.g. '/etc/51degrees-mobile-detector.settings.py') and module
names (e.g. 'myproject.fiftyone_degrees_mobile_settings') are allowed.
If not specified, defaults to '51degrees-mobile-detector.settings.py'
in the current working directory.

Additionally, when using the mobile detector in a Django project, settings
can also be specified using the FIFTYONE_DEGREES_MOBILE_DETECTOR_SETTINGS
variable in the Django settings file.

:copyright: (c) 2015 by 51Degrees.com, see README.md for more details.
:license: MPL2, see LICENSE.txt for more details.
'''

from __future__ import absolute_import
import os
import sys
import imp
import logging
from fiftyone_degrees.mobile_detector.conf import default


class _Settings(object):
    VERSION = '1.1'

    def __init__(self, settings_file_or_module):
        # Add default settings.
        self._add_settings(default)

        # Try to load settings from file/module pointed by the
        # environment variable.
        try:
            __import__(settings_file_or_module)
            self._add_settings(sys.modules[settings_file_or_module])
        except:
            try:
                self._add_settings(imp.load_source(
                    'fiftyone_degrees.conf._file',
                    settings_file_or_module))
            except:
                pass

        # Try to load setting from the Django settings file.
        try:
            from django.conf import settings
            from django.core import exceptions
            try:
                for name, value in getattr(settings, 'FIFTYONE_DEGREES_MOBILE_DETECTOR_SETTINGS', {}).iteritems():
                    self._add_setting(name, value)
            except exceptions.ImproperlyConfigured:
                pass
        except ImportError:
            pass

        # Add logger instance.
        self.logger = logging.getLogger('fiftyone_degrees.mobile_detector')

    def _add_settings(self, mod):
        '''Updates this dict with mod settings (only ALL_CAPS).

        '''
        for name in dir(mod):
            if name == name.upper():
                self._add_setting(name, getattr(mod, name))

    def _add_setting(self, name, value):
        '''Updates this dict with a specific setting.

        '''
        if name == 'USAGE_SHARER_IGNORED_HEADER_FIELD_VALUES':
            value = tuple([item.upper() for item in value])
        setattr(self, name, value)

settings = _Settings(
    os.environ.get(
        'FIFTYONE_DEGREES_MOBILE_DETECTOR_SETTINGS',
        os.path.join(os.getcwd(), '51degrees-mobile-detector.settings.py')))



# -*- coding: utf-8 -*-

'''
Default 51Degrees Mobile Detector settings. Override these using the
FIFTYONE_DEGREES_MOBILE_DETECTOR_SETTINGS environment variable. Both file
paths (e.g. '/etc/51Degrees/51degrees-mobile-detector.settings.py') and module
names (e.g. 'myproject.fiftyone_degrees_mobile_settings') are allowed.
If not specified, defaults to '51degrees-mobile-detector.settings.py'
in the current working directory.

Note that when using the mobile detector in a Django project, settings
can also be specified using the FIFTYONE_DEGREES_MOBILE_DETECTOR_SETTINGS
variable in the Django settings file.

:copyright: (c) 2015 by 51Degrees.com, see README.md for more details.
:license: MPL2, see LICENSE.txt for more details.
'''

from __future__ import absolute_import
import os

###############################################################################
## GENERAL SETTINGS.
###############################################################################

# Sets the preferred mobile device detection method. Available options are:
#
#   - 'v3-wrapper': Requires '51degrees-mobile-detector-v3-wrapper' package.
#   - 'v3-trie-wrapper': Requires '51degrees-mobile-detector-v3-trie-wrapper' package.
#
DETECTION_METHOD = 'v3-wrapper'

# List of case-sensitive property names to be fetched on every device detection. Leave empty to
# fetch all available properties.
PROPERTIES = ''
# Your 51Degrees license key. This is required if you want to set up the automatic
# data file updates.
LICENSE = ''

###############################################################################
## TRIE DETECTOR SETTINGS.
###############################################################################

# Location of the database file. If not specified, the trie-based detection
# method will not be available. Download the latest 51Degrees-LiteV3.2.trie
# file from http://github.com/51Degrees/Device-Detection/data/.
# Compare database options at https://51degrees.com/compare-data-options .
V3_TRIE_WRAPPER_DATABASE = os.path.expanduser('~/51Degrees/51Degrees-LiteV3.2.trie')

###############################################################################
## PATTERN DETECTOR SETTINGS.
###############################################################################

# Location of the database file. If not specified, the trie-based detection
# method will not be available. Download the latest 51Degrees-LiteV3.2.dat
# file from http://github.com/51Degrees/Device-Detection/data/.
# Compare database options at https://51degrees.com/compare-data-options .
V3_WRAPPER_DATABASE = os.path.expanduser('~/51Degrees/51Degrees-LiteV3.2.dat')

# Size of cache allocated
CACHE_SIZE = 10000

#Size of pool allocated
POOL_SIZE = 20

###############################################################################
## USAGE SHARER SETTINGS.
###############################################################################

# Indicates if usage data should be shared with 51Degrees.com. We recommended
# leaving this value unchanged to ensure we're improving the performance and
# accuracy of the solution.
USAGE_SHARER_ENABLED = True

# The detail that should be provided relating to new devices.
# Modification not required for most users.
USAGE_SHARER_MAXIMUM_DETAIL = True

# URL to send new device data to.
# Modification not required for most users.
USAGE_SHARER_SUBMISSION_URL = 'http://devices.51degrees.com/new.ashx'

# Data submission timeout (seconds).
USAGE_SHARER_SUBMISSION_TIMEOUT = 10

# Minimum queue length to launch data submission.
USAGE_SHARER_MINIMUM_QUEUE_LENGTH = 50

# Used to detect local devices.
# Modification not required for most users.
USAGE_SHARER_LOCAL_ADDRESSES = (
    '127.0.0.1',
    '0:0:0:0:0:0:0:1',
)

# The content of fields in this list should not be included in the
# request information sent to 51Degrees.
# Modification not required for most users.
USAGE_SHARER_IGNORED_HEADER_FIELD_VALUES = (
    'Referer',
    'cookie',
    'AspFilterSessionId',
    'Akamai-Origin-Hop',
    'Cache-Control',
    'Cneonction',
    'Connection',
    'Content-Filter-Helper',
    'Content-Length',
    'Cookie',
    'Cookie2',
    'Date',
    'Etag',
    'If-Last-Modified',
    'If-Match',
    'If-Modified-Since',
    'If-None-Match',
    'If-Range',
    'If-Unmodified-Since',
    'IMof-dified-Since',
    'INof-ne-Match',
    'Keep-Alive',
    'Max-Forwards',
    'mmd5',
    'nnCoection',
    'Origin',
    'ORIGINAL-REQUEST',
    'Original-Url',
    'Pragma',
    'Proxy-Connection',
    'Range',
    'Referrer',
    'Script-Url',
    'Unless-Modified-Since',
    'URL',
    'UrlID',
    'URLSCAN-ORIGINAL-URL',
    'UVISS-Referer',
    'X-ARR-LOG-ID',
    'X-Cachebuster',
    'X-Discard',
    'X-dotDefender-first-line',
    'X-DRUTT-REQUEST-ID',
    'X-Initial-Url',
    'X-Original-URL',
    'X-PageView',
    'X-REQUEST-URI',
    'X-REWRITE-URL',
    'x-tag',
    'x-up-subno',
    'X-Varnish',
)



# -*- coding: utf-8 -*-

'''
:copyright: (c) 2015 by 51Degrees.com, see README.md for more details.
:license: MPL2, see LICENSE.txt for more details.
'''

from __future__ import absolute_import
from abc import ABCMeta
from fiftyone_degrees.mobile_detector.conf import settings
from fiftyone_degrees.mobile_detector import usage

class Device(object):
    '''Simple device wrapper.

    '''
    def __init__(self, method=None):
        self._method = method
        self._properties = {}

    def set_property(self, name, value):
        self._properties[name] = value

    @property
    def method(self):
        return self._method

    @property
    def properties(self):
        return self._properties

    def __getattr__(self, name):
        if name in self._properties:
            return self._properties.get(name)
        else:
            name = name.lower()
            for aname, value in self._properties.iteritems():
                if name == aname.lower():
                    return value
        return None
    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)


class _Matcher(object):
    '''Abstract matcher class.

    '''
    __metaclass__ = ABCMeta
    _METHODS = {}
    _INSTANCES = {}

    @classmethod
    def register(cls, method, klass):
        cls._METHODS[method] = klass

    @classmethod
    def instance(cls, method):
        if method in cls._METHODS:
            if method not in cls._INSTANCES:
                cls._INSTANCES[method] = cls._METHODS[method]()
            return cls._INSTANCES[method]
        else:
            raise Exception(
                'Requested matching method "%s" does not exist. '
                'Available methods are: %s.' %
                (method, ', '.join(cls._METHODS.keys()),))

    def match(self, user_agent, client_ip=None, http_headers=None):
        # If provided, share usage information.
        if client_ip and http_headers:
            usage.UsageSharer.instance().record(client_ip, http_headers)

        # Delegate on specific matcher implementation.
        return self._match(user_agent)

    def _match(self, user_agent):
        raise NotImplementedError('Please implement this method.')


class _V3WrapperMatcher(_Matcher):
    ID = 'v3-wrapper'

    def __init__(self):
        if settings.V3_WRAPPER_DATABASE:
            try:
                # Does the database file exists and is it readable?
                with open(settings.V3_WRAPPER_DATABASE):
                    pass
            except IOError:
                raise Exception(
                    'The provided detection database file (%s) does not '
                    'exist or is not readable. Please, '
                    'check your settings.' % settings.V3_WRAPPER_DATABASE)
            else:
		from FiftyOneDegrees import fiftyone_degrees_mobile_detector_v3_wrapper
		self.provider = fiftyone_degrees_mobile_detector_v3_wrapper.Provider(settings.V3_WRAPPER_DATABASE, settings.PROPERTIES, int(settings.CACHE_SIZE), int(settings.POOL_SIZE))
        else:
            raise Exception(
                'Trie-based detection method depends on an external '
                'database file. Please, check your settings.')

    def _match(self, user_agent):
        # Delegate on wrapped implementation.
        returnedMatch = None
        try:
            returnedMatch = self.provider.getMatch(user_agent)
        except Exception as e:
            settings.logger.error(
                'Got exception while matching user agent string "%s": %s.'
                % (user_agent, unicode(e),))

        # Pythonize result.
        result = Device(self.ID)

	if returnedMatch:
		result.set_property('Id', returnedMatch.getDeviceId())
		result.set_property('MatchMethod', returnedMatch.getMethod())
		result.set_property('Difference', returnedMatch.getDifference())
		result.set_property('Rank', returnedMatch.getRank())
		if settings.PROPERTIES == '':
			for key in self.provider.getAvailableProperties():
				value = returnedMatch.getValues(key)
				if value:
					result.set_property(key, ' '.join(value))
				else:
					result.set_property(key, 'N/A in Lite')
		else:
			for key in settings.PROPERTIES.split(','):
				value = returnedMatch.getValues(key)
				if value:
					result.set_property(key, ' '.join(value))

        # Done!
        return result



class _V3TrieWrapperMatcher(_Matcher):
    ID = 'v3-trie-wrapper'

    def __init__(self):
        if settings.V3_TRIE_WRAPPER_DATABASE:
            try:
                # Does the database file exists and is it readable?
                with open(settings.V3_TRIE_WRAPPER_DATABASE):
                    pass
            except IOError:
                raise Exception(
                    'The provided detection database file (%s) does not '
                    'exist or is not readable. Please, '
                    'check your settings.' % settings.V3_TRIE_WRAPPER_DATABASE)
            else:
		from FiftyOneDegrees import fiftyone_degrees_mobile_detector_v3_trie_wrapper
		self.provider = fiftyone_degrees_mobile_detector_v3_trie_wrapper.Provider(settings.V3_TRIE_WRAPPER_DATABASE, settings.PROPERTIES)
        else:
            raise Exception(
                'Trie-based detection method depends on an external '
                'database file. Please, check your settings.')

    def _match(self, user_agent):
        # Delegate on wrapped implementation.
        returnedMatch = None
        try:
            returnedMatch = self.provider.getMatch(user_agent)
        except Exception as e:
            settings.logger.error(
                'Got exception while matching user agent string "%s": %s.'
                % (user_agent, unicode(e),))

        # Pythonize result.
        result = Device(self.ID)

	print settings.PROPERTIES
	if returnedMatch:
		if settings.PROPERTIES == '':
			for key in self.provider.getAvailableProperties():
				value = returnedMatch.getValues(key)
				if value:
					result.set_property(key, ' '.join(value))
				else:
					result.set_property(key, 'N/A in Lite')
		else:
			for key in settings.PROPERTIES.split(','):
				value = returnedMatch.getValues(key)
				if value:
					result.set_property(key, ' '.join(value))

        # Done!
        return result




# Register matching methods.
for klass in [_V3WrapperMatcher, _V3TrieWrapperMatcher]:
    _Matcher.register(klass.ID, klass)


def match(user_agent, client_ip=None, http_headers=None, method=None):
    '''Fetches device data based on an user agent string.

        *user_agent* is an user agent string.

        *client_ip* is a string with the client IP address (optional). If provided
        it'll will be submitted to 51Degrees.mobi in order to improve performance
        and accuracy of further device detections.

        *http_headers* is a dictionary containing all HTTP headers (optional).
        If provided, it'll will be submitted (removing confidential data such as
        cookies) to 51Degrees.mobi in order to improve performance and accuracy
        of further device detections.

        *method* is a string with the desired device detection method. If not
        specified, settings.DETECTION_METHOD will be used.

        Returns Device instance.
    '''
    # Fetch matcher instance.
    matcher = _Matcher.instance(
        method
        if method is not None
        else settings.DETECTION_METHOD)

    # Match!
    return matcher.match(user_agent, client_ip, http_headers)




# -*- coding: utf-8 -*-

'''
:copyright: (c) 2015 by 51Degrees.com, see README.md for more details.
:license: MPL2, see LICENSE.txt for more details.
'''

from __future__ import absolute_import
from datetime import datetime
import gzip
import urllib2
import threading
try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET
try:
    import cStringIO as StringIO
except ImportError:
    import StringIO
from fiftyone_degrees.mobile_detector.conf import settings


class UsageSharer(threading.Thread):
    '''Class used to record usage information.

    Records & submits usage information which is essential to ensuring
    51Degrees.mobi is optimized for performance and accuracy for current
    devices on the web.

    '''
    # Singleton reference.
    _instance = None

    def __init__(self):
        super(UsageSharer, self).__init__()

        # Check singleton instance.
        if self._instance is not None:
            raise ValueError('An instance of UsageSharer class already exists.')

        # Init internal singleton state.
        self._queue = []
        self._stopping = False
        self._event = threading.Event()

    @classmethod
    def instance(cls):
        '''Returns singleton UsageSharer instance.

        '''
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def record(self, client_ip, http_headers):
        '''Adds request details.

        Adds some request details to the queue for further submission by
        the background thread.

        *client_ip* is a string with the client IP address.

        *http_headers* is a dictionary containing all HTTP headers.

        '''
        # Enabled?
        if settings.USAGE_SHARER_ENABLED:
            # Launch background daemon data submission thread if not running.
            if not self.is_alive():
                self.daemon = True
                self.start()

            # Add the request details to the queue for further submission.
            self._queue.append(self._get_item(client_ip, http_headers))

            # Signal the background thread to check if it should send
            # queued data.
            self._event.set()

    def stop(self):
        '''Gracefully stops background data submission thread.

        '''
        if self.is_alive():
            settings.logger.info('Stopping 51Degrees.mobi UsageSharer.')
            self._stopping = True
            self._event.set()
            self.join()

    def run(self):
        '''Runs the background daemon data submission thread.

        Used to send the devices data back to 51Degrees.mobi after the
        minimum queue length has been reached.

        '''
        # Log.
        settings.logger.info('Starting 51Degrees.mobi UsageSharer.')

        # Submission loop.
        while not self._stopping:
            # Wait while event's flag is set to True.
            while not self._event.is_set():
                self._event.wait()

            # If there are enough items in the queue, or the thread is being
            # stopped, submit the queued data.
            length = len(self._queue)
            if length >= settings.USAGE_SHARER_MINIMUM_QUEUE_LENGTH or (length > 0 and self._stopping):
                self._submit()

            # Reset the internal event's flag to False.
            self._event.clear()

        # Log.
        settings.logger.info('Stopped 51Degrees.mobi UsageSharer.')

    def _is_local(self, address):
        return address in settings.USAGE_SHARER_LOCAL_ADDRESSES

    def _get_item(self, client_ip, http_headers):
        # Create base device element.
        device = ET.Element('Device')

        # Add the current date and time.
        item = ET.SubElement(device, 'DateSent')
        item.text = datetime.utcnow().replace(microsecond=0).isoformat()

        # Add product name and version.
        item = ET.SubElement(device, 'Version')
        item.text = settings.VERSION
        item = ET.SubElement(device, 'Product')
        item.text = 'Python Mobile Detector'

        # Add client IP address (if is not local).
        if not self._is_local(client_ip):
            item = ET.SubElement(device, 'ClientIP')
            item.text = client_ip

        # Filter & add HTTP headers.
        for name, value in http_headers.iteritems():
            # Determine if the field should be treated as a blank.
            blank = name.upper() in settings.USAGE_SHARER_IGNORED_HEADER_FIELD_VALUES

            # Include all header values if maximum detail is enabled, or
            # header values related to the user agent or any header
            # key containing profile or information helpful to determining
            # mobile devices.
            if settings.USAGE_SHARER_MAXIMUM_DETAIL or \
               name.upper() in ('USER-AGENT', 'HOST', 'PROFILE') or \
               blank:
                item = ET.SubElement(device, 'Header')
                item.set('Name', name)
                if not blank:
                    item.text = unicode(value)

        # Done!
        return device

    def _submit(self):
        '''Sends all the data on the queue.

        '''
        settings.logger.info('Submitting UsageSharer queued data to 51Degrees.mobi.')

        # Build output stream.
        stream = StringIO.StringIO()
        devices = ET.Element('Devices')
        while len(self._queue) > 0:
            devices.append(self._queue.pop())
        ET.ElementTree(devices).write(
            stream,
            encoding='utf8',
            xml_declaration=True)

        # Submit gzipped data.
        request = urllib2.Request(
            url=settings.USAGE_SHARER_SUBMISSION_URL,
            data=gzip.GzipFile(fileobj=stream).read(),
            headers={
                'Content-Type': 'text/xml; charset=utf-8',
                'Content-Encoding': 'gzip',
            })
        try:
            response = urllib2.urlopen(request, timeout=settings.USAGE_SHARER_SUBMISSION_TIMEOUT)
        except:
            # Turn off functionality.
            self._stopping = True
        else:
            # Get the response and record the content if it's valid. If it's
            # not valid consider turning off the functionality.
            code = response.getcode()
            if code == 200:
                # OK. Do nothing.
                pass
            elif code == 408:
                # Request Timeout. Could be temporary, do nothing.
                pass
            else:
                # Turn off functionality.
                self._stopping = True
